<?php

namespace Drupal\pdf_using_mpdf\Conversion;

use Drupal\Core\Asset\LibraryDiscoveryCollector;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\File\FileUrlGenerator;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Theme\ThemeManagerInterface;
use Drupal\Core\Utility\Token;
use Drupal\file\FileRepositoryInterface;
use Drupal\pdf_using_mpdf\ConvertToPdfInterface;
use Mpdf\HTMLParserMode;
use Mpdf\Mpdf;
use Mpdf\MpdfException;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

/**
 * HTML to PDF converter service.
 */
class ConvertToPdf implements ConvertToPdfInterface {
  use StringTranslationTrait;
  /**
   * Renderer service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * The Mpdf object.
   *
   * @var \Mpdf\Mpdf
   */
  protected $mpdf;

  /**
   * Token object.
   *
   * @var \Drupal\Core\Utility\Token
   */
  protected $token;

  /**
   * Context data.
   *
   * @var array
   */
  protected $context;

  /**
   * Configuration object.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The entityTypeManager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The file repository service.
   *
   * @var \Drupal\file\FileRepositoryInterface
   */
  protected $fileRepository;


  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * File url generator object.
   *
   * @var \Drupal\Core\File\FileUrlGenerator
   */
  protected $fileUrlGenerator;

  /**
   * Logger object.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $logger;

  /**
   * The Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The Drupal Theme Manager service.
   *
   * @var \Drupal\Core\Theme\ThemeManagerInterface
   */
  protected $themeManager;

  /**
   * The Drupal Asset Library Discovery Service.
   *
   * @var Drupal\Core\Asset\LibraryDiscoveryCollector
   */
  protected $libraryDiscovery;

  /**
   * Pdf_using_mpdf settings.
   *
   * @var array
   */
  protected $settings;

  /**
   * Hold all the loaded CSS files from theme.
   *
   * @var string
   */
  protected $allCssFiles;

  /**
   * ConvertToPdf constructor.
   *
   * @param \Drupal\Core\Render\RendererInterface $renderer
   *   Renderer service definition.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   Config factory service definition.
   * @param \Drupal\Core\Utility\Token $token
   *   Token service definition.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger
   *   Logger service definition.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system service.
   * @param \Drupal\file\FileRepositoryInterface $file_repository
   *   The file repository service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\File\FileUrlGenerator $fileUrlGenerator
   *   File url generator object.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   * @param \Drupal\Core\Theme\ThemeManagerInterface $theme_manager
   *   The theme manager service.
   * @param Drupal\Core\Asset\LibraryDiscoveryCollector $library_discovery
   *   The theme manager service.
   */
  public function __construct(RendererInterface $renderer, ConfigFactoryInterface $config_factory, Token $token, LoggerChannelFactoryInterface $logger, FileSystemInterface $file_system, FileRepositoryInterface $file_repository, EntityTypeManagerInterface $entity_type_manager, FileUrlGenerator $fileUrlGenerator, MessengerInterface $messenger, ThemeManagerInterface $theme_manager, LibraryDiscoveryCollector $library_discovery) {
    $this->renderer = $renderer;
    $this->configFactory = $config_factory;
    $this->token = $token;
    $this->logger = $logger;
    $this->fileSystem = $file_system;
    $this->fileRepository = $file_repository;
    $this->entityTypeManager = $entity_type_manager;
    $this->fileUrlGenerator = $fileUrlGenerator;
    $this->messenger = $messenger;
    $this->themeManager = $theme_manager;
    $this->libraryDiscovery = $library_discovery;

    // Initialize with settings from db.
    $this->settings = $this->getConfig();
    $this->allCssFiles = '';
  }

  /**
   * {@inheritdoc}
   */
  public function convert($html, $settings = [], $options = []) {
    if (empty($html)) {
      $this->logger
        ->get('pdf_using_mpdf')
        ->error($this->t('There was an error generating PDF - No HTML content provided.'));

      return;
    }

    // Overwrite settings - altered via hook_mpdf_settings_alter()
    $this->settings = array_merge($this->settings, $this->getDefaultConfig(), $settings);
    $this->context = $options;
    return $this->generator($html);
  }

  /**
   * Generates PDF file from the HTML.
   *
   * @param string $html
   *   Contents of the template already with the node data.
   */
  protected function generator($html) {
    $this->importStyles();
    $this->replaceAllSettingsTokens();

    try {
      $this->mpdf = new Mpdf($this->settings);
      $this->setHeader();
      $this->applyProperties();
      $this->setFooter();

      // Apply custom cascading styles.
      if (!empty($this->allCssFiles)) {
        $this->mpdf->WriteHTML($this->allCssFiles, HTMLParserMode::HEADER_CSS);
      }

      $backtrack_limit = (int) ini_get('pcre.backtrack_limit');
      if (!empty($backtrack_limit) && strlen($html) > $backtrack_limit) {
        $chunk_len = $backtrack_limit / 10;
        $chunk_str = chunk_split($html, $chunk_len, '~#~~~#~');
        $html_chunks = array_filter(explode('~#~~~#~', $chunk_str));
        foreach ($html_chunks as $chunk) {
          $this->mpdf->WriteHTML($chunk, HTMLParserMode::HTML_BODY);
        }
      }
      else {
        $this->mpdf->WriteHTML($html, HTMLParserMode::HTML_BODY);
      }

      return $this->output();
    }
    catch (MpdfException $e) {
      $message = $e->getMessage();
      $this->logger->get('pdf_using_mpdf')->critical('Error:' . $message);

      return;
    }
  }

  /**
   * Get configuration from database.
   *
   * @return array
   *   Stored configuration values.
   */
  public function getConfig() {
    return $this->configFactory
      ->getEditable('pdf_using_mpdf.settings')
      ->get('pdf_using_mpdf');
  }

  /**
   * Set header for PDF file.
   */
  protected function setHeader() {
    $header = $this->settings['pdf_header'];

    if (isset($header) && $header != NULL) {
      $this->mpdf->SetHTMLHeader($header);
    }
  }

  /**
   * Apply additional properties to PDF file.
   */
  protected function applyProperties() {

    // Set Watermark.
    $watermark_option = $this->settings['watermark_option'];
    $watermark_opacity = $this->settings['watermark_opacity'];

    if ($watermark_option == 0) {
      $text = $this->settings['pdf_watermark_text'];
      if (!empty($text)) {
        $this->mpdf->SetWatermarkText($text, $watermark_opacity);
        $this->mpdf->showWatermarkText = TRUE;
      }
    }
    else {
      $image_id = $this->settings['watermark_image'];
      if (isset($image_id[0])) {
        $file = $this->entityTypeManager->getStorage('file')->load($image_id[0]);
        $image_path = $file->getFileUri();
        $image_path = $this->fileUrlGenerator->generateAbsoluteString($image_path);
        $this->mpdf->SetWatermarkImage($image_path, $watermark_opacity);
        $this->mpdf->showWatermarkImage = TRUE;
      }
    }

    // Set Title.
    $title = $this->settings['pdf_set_title'];
    if (!empty($title)) {
      $this->mpdf->SetTitle($title);
    }

    // Set Author.
    $author = $this->settings['pdf_set_author'];
    if (!empty($author)) {
      $this->mpdf->SetAuthor($author);
    }

    // Set Subject.
    $subject = $this->settings['pdf_set_subject'];
    if (isset($subject) && $subject != NULL) {
      $this->mpdf->SetSubject($subject);
    }

    // Set Creator.
    $creator = $this->settings['pdf_set_creator'];
    if (!empty($creator)) {
      $this->mpdf->SetCreator($creator);
    }

    // Set Password.
    $password = $this->settings['pdf_password'];
    if (!empty($password)) {
      $this->mpdf->SetProtection(['print', 'copy'], $password, $password);
    }
  }

  /**
   * Set footer for PDF file.
   */
  protected function setFooter() {
    $footer = $this->settings['pdf_footer'];
    if (isset($footer) && $footer != NULL) {
      $this->mpdf->SetHTMLFooter($footer);
    }
  }

  /**
   * Output PDF file.
   */
  public function output() {
    $filename = $this->settings['pdf_filename'];
    switch ($this->settings['pdf_save_option']) {
      case 0:
        // Web browser.
        try {
          // Set (if not already) title to show on the browser as file name.
          if (empty($this->settings['pdf_set_title'])) {
            $this->mpdf->SetTitle($filename . '.pdf');
          }

          // Get content.
          $pdfContent = $this->mpdf->OutputBinaryData();

          // Prepare and send response.
          $response = new Response($pdfContent);
          $response->headers->set('Content-Type', 'application/pdf');
          $response->headers->set('Content-disposition', 'inline; filename=' . $filename . '.pdf');
          return $response;

        }
        catch (MpdfException $e) {
          $message = 'Web browser: PDF file generation error. ' . $e->getMessage();
          $this->logger->get('pdf_using_mpdf')->critical($message);
          throw new BadRequestHttpException($message);
        }
        break;

      case 1:
        // Save Dialog box.
        try {
          // Get content.
          $pdfContent = $this->mpdf->OutputBinaryData();

          // Prepare and send response.
          $response = new Response($pdfContent);
          $response->headers->set('Content-Description', 'File Transfer');
          $response->headers->set('Content-Transfer-Encoding', 'Transfer-Encoding');
          $response->headers->set('Content-Type', 'application/pdf');
          $response->headers->set('Content-disposition', 'attachment; filename=' . $filename . '.pdf');
          return $response;

        }
        catch (MpdfException $e) {
          $message = 'Download: PDF file generation error. ' . $e->getMessage();
          $this->logger->get('pdf_using_mpdf')->critical($message);
          throw new BadRequestHttpException($message);
        }
        break;

      case 2:
        // Save to server.
        try {

          // Get content.
          $pdfContent = $this->mpdf->OutputBinaryData();

          // Get schema and path.
          $scheme = $this->configFactory
            ->getEditable('system.file')
            ->get('default_scheme');
          $path = $scheme . '://pdf_using_mpdf/';
          $fullpath = $path . $filename . '.pdf';
          $realpath = $this->fileSystem->realpath($fullpath);

          // Prepare dir and create file.
          $this->fileSystem->prepareDirectory($path, FileSystemInterface::CREATE_DIRECTORY);
          $this->fileRepository->writeData($pdfContent, $fullpath);

          $this->messenger->addStatus($this->t('PDF file %filename saved to %path', [
            '%filename' => $filename . '.pdf',
            '%path' => $realpath,
          ]));

          return [];
        }
        catch (MpdfException $e) {
          $message = 'Save to server: PDF file generation error. ' . $e->getMessage();
          $this->logger->get('pdf_using_mpdf')->critical($message);
          throw new BadRequestHttpException($message);
        }
        break;
    }
  }

  /**
   * Configuration values to instantiate Mpdf constructor.
   *
   * @return array
   *   Default configs array.
   */
  public function getDefaultConfig() {
    $orientation = $this->settings['orientation'] == 'L' ? '-L' : '';

    return [
      'tempDir' => $this->fileSystem->getTempDirectory(),
      'useActiveForms' => TRUE,
      'format' => $this->settings['pdf_page_size'] . $orientation,
      'default_font_size' => $this->settings['pdf_font_size'],
      'default_font' => $this->settings['pdf_default_font'],
      'margin_left' => $this->settings['margin_left'],
      'margin_right' => $this->settings['margin_right'],
      'margin_top' => $this->settings['margin_top'],
      'margin_bottom' => $this->settings['margin_bottom'],
      'margin_header' => $this->settings['margin_header'],
      'margin_footer' => $this->settings['margin_footer'],
      'dpi' => $this->settings['dpi'],
      'img_dpi' => $this->settings['img_dpi'],
    ];
  }

  /**
   * Import CSS files from file or call recursively..
   */
  protected function importStyles() {
    if (isset($this->settings['pdf_css_file']) && !empty($this->settings['pdf_css_file'])) {
      $path = DRUPAL_ROOT . '/' . $this->settings['pdf_css_file'];
      if (($this->fileSystem->realpath($path) === $path) && (file_exists($path))) {
        $this->allCssFiles .= file_get_contents($path);
      }
    }
    elseif ($this->settings['pdf_css_from_theme']) {

      if ($this->settings['pdf_css_from_theme_all']) {
        // Load all libraries from theme.
        $libraries = $this->libraryDiscovery->getLibrariesByExtension($this->themeManager->getActiveTheme()->getName());

        foreach ($libraries as $data) {
          if (isset($data['css'])) {
            foreach ($data['css'] as $item) {
              if ($item['type'] == 'file') {
                $file = $item['data'];
                $absolute_path = $this->fileSystem->realpath($file);
                $this->allCssFiles .= file_get_contents($absolute_path);
              }
            }
          }
        }
      }
      else {
        // Get current libraries from theme.
        $libraries = $this->themeManager
          ->getActiveTheme()
          ->getLibraries();

        foreach ($libraries as $library) {
          [$ext, $name] = explode('/', $library);
          $data = $this->libraryDiscovery->getLibraryByName($ext, $name);
          if (isset($data['css'])) {
            foreach ($data['css'] as $item) {
              if ($item['type'] == 'file') {
                $file = $item['data'];
                $absolute_path = $this->fileSystem->realpath($file);
                $this->allCssFiles .= file_get_contents($absolute_path);
              }
            }
          }
        }
      }
    }
  }

  /**
   * Replace all tokens.
   */
  public function replaceAllSettingsTokens() {
    $settings = [];
    foreach ($this->settings as $key => $value) {
      $settings[$key] = !is_array($value) ? $this->token->replace($value, $this->context) : $value;
    }

    $this->settings = $settings;
  }

}
